# tessif/frused/hooks/__init__.py
# -*- coding: utf-8 -*-
"""
:mod:`~tessif.frused.hooks` is a :mod:`tessif` subpackage aggregating
:ref:`model specific <SupportedModels>` hooks to improve their integration
into tessif. Collecting them all in one place enables convenient external
use as well as easy maintenance.

:mod:`~tessif.frused.hooks` serves as main reference point for adjusting
model specific integration behavior.
"""
